@@ -38,7 +38,7 @@ typedef struct {
 
 	SoupMessage *cur_req;
 	SoupConnectionState state;
-	time_t       last_used;
+	gboolean     ever_used;
 	guint        io_timeout, idle_timeout;
 	GSource     *idle_timeout_src;
 } SoupConnectionPrivate;
@@ -90,6 +90,8 @@ finalize (GObject *object)
 		g_object_unref (priv->remote_addr);
 	if (priv->tunnel_addr)
 		g_object_unref (priv->tunnel_addr);
+	if (priv->proxy_uri)
+		soup_uri_free (priv->proxy_uri);
 
 	if (priv->async_context)
 		g_main_context_unref (priv->async_context);
@@ -335,7 +337,7 @@ clear_current_request (SoupConnection *conn)
 		if (!soup_message_is_keepalive (cur_req))
 			soup_connection_disconnect (conn);
 		else {
-			priv->last_used = time (NULL);
+			priv->ever_used = TRUE;
 			soup_message_io_stop (cur_req);
 		}
 	}
@@ -537,10 +539,10 @@ soup_connection_disconnect (SoupConnection *conn)
 
 	if (priv->cur_req &&
 	    priv->cur_req->status_code == SOUP_STATUS_IO_ERROR &&
-	    priv->last_used != 0) {
+	    priv->ever_used) {
 		/* There was a message queued on this connection, but
 		 * the socket was closed while it was being sent.
-		 * Since last_used is not 0, then that means at least
+		 * Since ever_used is TRUE, then that means at least
 		 * one message was successfully sent on this
 		 * connection before, and so the most likely cause of
 		 * the IO_ERROR is that the connection was idle for
@@ -559,14 +561,14 @@ soup_connection_disconnect (SoupConnection *conn)
 					    SOUP_MESSAGE_IO_STATUS_QUEUED);
 	}
 
-	/* If cur_req is non-NULL but priv->last_used is 0, then that
+	/* If cur_req is non-NULL but priv->ever_used is FALSE, then that
 	 * means this was the first message to be sent on this
 	 * connection, and it failed, so the error probably means that
 	 * there's some network or server problem, so we let the
 	 * IO_ERROR be returned to the caller.
 	 *
 	 * (Of course, it's also possible that the error in the
-	 * last_used != 0 case was because of a network/server problem
+	 * ever_used == TRUE case was because of a network/server problem
 	 * too. It's even possible that the message crashed the
 	 * server. In this case, requeuing it was the wrong thing to
 	 * do, but presumably, the next attempt will also get an
@@ -598,10 +600,25 @@ soup_connection_get_proxy_uri (SoupConnection *conn)
 SoupConnectionState
 soup_connection_get_state (SoupConnection *conn)
 {
+	SoupConnectionPrivate *priv;
+
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn),
 			      SOUP_CONNECTION_DISCONNECTED);
+	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	return SOUP_CONNECTION_GET_PRIVATE (conn)->state;
+#ifdef G_OS_UNIX
+	if (priv->state == SOUP_CONNECTION_IDLE) {
+		GPollFD pfd;
+
+		pfd.fd = soup_socket_get_fd (priv->socket);
+		pfd.events = G_IO_IN;
+		pfd.revents = 0;
+		if (g_poll (&pfd, 1, 0) == 1)
+			priv->state = SOUP_CONNECTION_REMOTE_DISCONNECTED;
+	}
+#endif
+
+	return priv->state;
 }
 
 void
@@ -616,23 +633,6 @@ soup_connection_set_state (SoupConnection *conn, SoupConnectionState state)
 		clear_current_request (conn);
 }
 
-/**
- * soup_connection_last_used:
- * @conn: a #SoupConnection.
- *
- * Returns the last time a response was received on @conn.
- *
- * Return value: the last time a response was received on @conn, or 0
- * if @conn has not been used yet.
- */
-time_t
-soup_connection_last_used (SoupConnection *conn)
-{
-	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
-
-	return SOUP_CONNECTION_GET_PRIVATE (conn)->last_used;
-}
-
 /**
  * soup_connection_send_request:
  * @conn: a #SoupConnection
